#include "Cmpl.hpp"
#include "../plugin.hpp"
#include "ECmplLng.h"
#include "MatPat.h"
#include <limits.h> // INT_MAX
#include <stdlib.h>
#if defined(__GNUC__)
#define strrev _strrev
#endif

TCompletion::TCompletion(void)
{
    Word[0] = 0;

    Stop = FALSE;
    WorkInsideWord = TRUE;
    CaseSensitive = FALSE;
    BrowseDownward = TRUE;
    AddTrailingSpace = FALSE;
    ConsiderDigitAsChar = TRUE;
    PartialCompletion = FALSE;
    MinPreWordLen = 1;
    MinWordLen = 2;
    BrowseLineCnt = 500;
    WordsToFindCnt = 10;
    SortListCount = 0;

    AdditionalLetters[0] = '_';
    AdditionalLetters[1] = 0;

    AsteriskSymbol = 0;

    RegKey[0] = 0;
    ConfigHelpTopic[0] = 0;

    WordCount = 0;
    WordList = NULL;
    DialogItemsCnt = CMPL_DIALOG_ITEMS;
    DialogWidth = 46;
    DialogHeight = 22;
};

TCompletion::~TCompletion()
{
    Word[0] = 0;

    WordCount = 0;
    WordList = NULL;
};

void TCompletion::GetOptions()
{
    WorkInsideWord      = GetRegKey( HKEY_CURRENT_USER, RegKey, "WorkInsideWord", WorkInsideWord );
    CaseSensitive       = GetRegKey( HKEY_CURRENT_USER, RegKey, "CaseSensitive", CaseSensitive );
    BrowseDownward      = GetRegKey( HKEY_CURRENT_USER, RegKey, "BrowseDownward", BrowseDownward );
    MinPreWordLen       = GetRegKey( HKEY_CURRENT_USER, RegKey, "MinPreWordLen", MinPreWordLen );
    MinWordLen          = GetRegKey( HKEY_CURRENT_USER, RegKey, "MinWordLen", MinWordLen );
    BrowseLineCnt       = GetRegKey( HKEY_CURRENT_USER, RegKey, "BrowseLineCnt", BrowseLineCnt );
    WordsToFindCnt      = GetRegKey( HKEY_CURRENT_USER, RegKey, "WordsToFindCnt", WordsToFindCnt );
    SortListCount       = GetRegKey( HKEY_CURRENT_USER, RegKey, "SortListCount", SortListCount );
    ConsiderDigitAsChar = GetRegKey( HKEY_CURRENT_USER, RegKey, "ConsiderDigitAsChar", ConsiderDigitAsChar );
    PartialCompletion   = GetRegKey( HKEY_CURRENT_USER, RegKey, "PartialCompletion", PartialCompletion );
    AddTrailingSpace    = GetRegKey( HKEY_CURRENT_USER, RegKey, "AddTrailingSpace", AddTrailingSpace );
    int DataSize = 80;
    GetRegKey( HKEY_CURRENT_USER, RegKey, "AdditionalLetters", AdditionalLetters, AdditionalLetters, DataSize );

    Validate( );
};

void TCompletion::SetOptions()
{
    Validate( );

    SetRegKey( HKEY_CURRENT_USER, RegKey, "WorkInsideWord", WorkInsideWord );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "CaseSensitive", CaseSensitive );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "BrowseDownward", BrowseDownward );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "BrowseLineCnt", BrowseLineCnt );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "MinPreWordLen", MinPreWordLen );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "MinWordLen", MinWordLen );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "WordsToFindCnt", WordsToFindCnt );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "SortListCount", SortListCount );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "ConsiderDigitAsChar", ConsiderDigitAsChar );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "PartialCompletion", PartialCompletion );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "AddTrailingSpace", AddTrailingSpace );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "AdditionalLetters", AdditionalLetters );
};

void TCompletion::Validate(void)
{
    WorkInsideWord      = WorkInsideWord ? TRUE : FALSE;
    CaseSensitive       = CaseSensitive ? TRUE : FALSE;
    BrowseDownward      = BrowseDownward ? TRUE : FALSE;
    PartialCompletion   = PartialCompletion ? TRUE : FALSE;
    AddTrailingSpace    = AddTrailingSpace ? TRUE : FALSE;
    ConsiderDigitAsChar = ConsiderDigitAsChar ? TRUE : FALSE;
    BrowseLineCnt       = BrowseLineCnt >= 0 ? BrowseLineCnt : INT_MAX;
    MinPreWordLen       = MAX( MinPreWordLen, 1 );
    MinPreWordLen       = MIN( MinPreWordLen, MAX_WORD_LEN-10 );
    MinWordLen          = MAX( MinWordLen, 2 );
    MinWordLen          = MIN( MinWordLen, MAX_WORD_LEN-10 );
    WordsToFindCnt      = MAX( WordsToFindCnt, 0 );
    WordsToFindCnt      = MIN( WordsToFindCnt, MAX_WORD_CNT );
    SortListCount       = MAX( SortListCount, 0 );
    SortListCount       = MIN( SortListCount, MAX_WORD_CNT-1 );
};


int TCompletion::GetCoincidenceLen(void)
{
    int Len = 0;

    if( WordCount && WordLen )
    {
        for( unsigned int i = WordLen; i <= strlen( WordList[0].Text ); i++ )
        {
            for( int j = 1; j < WordCount; j++ )
            {
                if( FSF.LStrnicmp( WordList[0].Text, WordList[j].Text, i ) )
                    goto _Exit;
            };
            Len = i;
        };
    };

_Exit:
    return Len;
};

BOOL TCompletion::DoPartialCompletion(void)
{
    BOOL WasCompleted = FALSE;

    if( PartialCompletion && WordList && WordCount && WordLen )
    {
        int CoincidenceLen = GetCoincidenceLen( );
        if( CoincidenceLen > WordLen )
        {
            WordList[0].Text[CoincidenceLen] = 0;
            WordCount = 1;
            WasCompleted = TRUE;
        };
    };

    return WasCompleted;
};

int TCompletion::GetPreWord(void)
{
    struct EditorInfo ei;
    struct EditorGetString gs;
    struct EditorConvertText ct;
    unsigned char *Line;

    Word[0] = 0;
    WordLen = 0;

    EditorControl(ECTL_GETINFO,&ei);

    gs.StringNumber=-1; // current string
    EditorControl(ECTL_GETSTRING,&gs);

    //        !!!
    if( ei.CurPos>0 && ei.CurPos <= gs.StringLength )
    {
        Line = (unsigned char *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, gs.StringLength+1);
        if( Line )
        {
            memcpy( Line, gs.StringText, gs.StringLength+1 );

            ct.TextLength = gs.StringLength;
            ct.Text = (char *)Line;
            EditorControl( ECTL_EDITORTOOEM, &ct );

            int Pos = ei.CurPos;

            //  WorkInsideWord==FALSE,       
            if( !MyIsAlpha(Line[Pos]) || WorkInsideWord )
            {
                //   
                while( (MyIsAlpha(Line[Pos-1]) && Pos) || (AsteriskSymbol && (Line[Pos-1]==AsteriskSymbol)) )
                    Pos--;

                if( Pos < ei.CurPos )
                {
                    WordPos = MyIsAlpha(Line[Pos]) ? 0 : 1;
                    WordPos += Pos;
                    WordLen = MIN(ei.CurPos-WordPos, MAX_WORD_LEN);
                    WordLen = MIN(WordLen, ct.TextLength);
                    strncpy( (char *)Word, (char *)Line+WordPos, WordLen );
                };
                Word[WordLen] = 0;
            };

            HeapFree(GetProcessHeap(), 0, Line);
        };
    };

    if( WordLen < MinPreWordLen )
    {
        Word[0] = 0;
        WordLen = 0;
    };

    return WordLen;
};

// ,  GetPreWord    Word 
//   
int TCompletion::DoSearch(void)
{
    struct EditorGetString gs;
    struct EditorInfo ei;
    struct EditorSetPosition sp;
    struct EditorConvertText ct;

    int i, j, LinCnt;
    int Line2Browse[2]; //       

    unsigned char *Line;

    WordCount = 0;
    MaxWordLen = 0;

    if( Word && Word[0] )
    {
        EditorControl(ECTL_GETINFO,&ei);

        gs.StringNumber=-1; // current string
        EditorControl(ECTL_GETSTRING,&gs);

        Line = (unsigned char *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, gs.StringLength+1);
        if( Line )
        {
            int StrMaxLen = gs.StringLength;

            memcpy( Line, gs.StringText, gs.StringLength+1 );

            ct.TextLength=gs.StringLength;
            ct.Text=(char *)Line;
            EditorControl(ECTL_EDITORTOOEM,&ct);

            //    
            AddWords( (char *)Line, ei.CurPos, 0 );

            //   -  ( Word)
            if( BrowseDownward )
                AddWords( (char *)&Line[ei.CurPos], gs.StringLength - ei.CurPos, 1 );

            Line2Browse[0]=MIN( ei.CurLine, BrowseLineCnt ); // 
            Line2Browse[1]=MIN( ei.TotalLines-ei.CurLine-1, BrowseLineCnt); // 

            LinCnt=MAX(Line2Browse[0], Line2Browse[1]); //    

            gs.StringNumber=-1; // . 
            sp.CurPos=-1;
            sp.CurTabPos=-1;
            sp.TopScreenLine=-1;
            sp.LeftPos=-1;
            sp.Overtype=-1;

            for( i=1; (i <= LinCnt) && (WordCount < WordsToFindCnt); i++ )
            {
                for( j=0; j < 1+BrowseDownward; j++ )
                {
                    if( Line2Browse[j] < i )
                        continue;
                    sp.CurLine=ei.CurLine+(j?i:-i); //  . 
                    EditorControl(ECTL_SETPOSITION,&sp);
                    EditorControl(ECTL_GETSTRING,&gs);

                    ct.TextLength = gs.StringLength;
                    if( StrMaxLen < ct.TextLength )
                    {
                        HeapFree( GetProcessHeap(), 0, Line );
                        StrMaxLen = 0;

                        SetLastError(0);
                        Line=(unsigned char *)HeapAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY, ct.TextLength+1 );
                        if( GetLastError() != 0 )
                        {
                            SetLastError(0);
                            continue;    //    -   
                        }
                        else
                            StrMaxLen = ct.TextLength;
                    };
                    memcpy( Line, gs.StringText, ct.TextLength+1);

                    ct.Text = (char *)Line;

                    EditorControl( ECTL_EDITORTOOEM, &ct );

                    AddWords( (char *)Line, ct.TextLength, j);

                    if( WordCount >= WordsToFindCnt )
                        break;
                };
                if(Stop) break;
            }

            sp.CurLine = ei.CurLine;
            sp.CurPos = ei.CurPos;
            sp.TopScreenLine = ei.TopScreenLine;
            sp.LeftPos = ei.LeftPos;
            EditorControl( ECTL_SETPOSITION, &sp );

            HeapFree( GetProcessHeap(), 0, Line );
        };
    };

    return WordCount;
};

BOOL TCompletion::MyIsAlpha(int c)
{
    BOOL ret = FALSE;

    if( c )
    {
        if( ConsiderDigitAsChar )
            ret = FSF.LIsAlphanum( c );
        else
            ret = FSF.LIsAlpha( c );
        if( !ret )
            if( strchr( AdditionalLetters, c ) )
                ret = TRUE;
    };

    return ret;
};

int CompareListItems(const void *a, const void *b)
{
    return FSF.LStricmp( (char *)(((struct FarMenuItem *)(a))->Text),
                         (char *)(((struct FarMenuItem *)(b))->Text) );
};

int CompareListItemsByFreq(const void *a, const void *b)
{
    return ( *((int *)(&(((struct FarMenuItem *)(b))->Text[MAX_WORD_LEN]))) -
             *((int *)(&(((struct FarMenuItem *)(a))->Text[MAX_WORD_LEN]))) );
};

void TCompletion::SortList(void)
{
    if( WordCount > SortListCount )
        qsort( WordList, WordCount, sizeof(struct FarMenuItem),
               CompareListItems );
    else
        qsort( WordList, WordCount, sizeof(struct FarMenuItem),
               CompareListItemsByFreq );

};

char *TCompletion::PutWord(void)
{
    struct EditorInfo ei;
    struct EditorGetString gs;

    static char OverwritedText[MAX_WORD_LEN];

    OverwritedText[0] = 0;

    EditorControl(ECTL_GETINFO,&ei);

    if( AddTrailingSpace && (WordLen <= (MAX_WORD_LEN - 2)) )
    {
        Word[WordLen++] = ' ';
        Word[WordLen] = 0;
    };

    SetCurPos( WordPos );

    if( !ei.Overtype )
    {
        for( int i = WordPos; i < ei.CurPos; i++ )
            EditorControl( ECTL_DELETECHAR, NULL );
    }
    else
    {
        gs.StringNumber = -1; // current string
        EditorControl( ECTL_GETSTRING, &gs );
        memcpy( OverwritedText, &gs.StringText[ei.CurPos], WordLen-ei.CurPos+WordPos );
        OverwritedText[WordLen-ei.CurPos+WordPos] = 0;
    };

    EditorControl( ECTL_INSERTTEXT, Word );

    SetCurPos( ei.CurPos );

    return OverwritedText;
};

void TCompletion::SetCurPos(int NewPos)
{
    struct EditorSetPosition sp;

    sp.CurLine = -1;
    sp.CurPos = NewPos; //WordPos + WordLen;
    sp.CurTabPos = -1;
    sp.TopScreenLine = -1;
    sp.LeftPos = -1;
    sp.Overtype = -1;
    Info.EditorControl( ECTL_SETPOSITION, &sp );
};

int TCompletion::AddWords(char *Line, int Len, int Direction)
{
    char CurCharStr[2];
    char TmpWord[MAX_WORD_LEN];
    BOOL IsInsideWord = FALSE;
    int i;
    unsigned char CurChar;

    int LineWordsCnt = 0;

    CurCharStr[1] = 0;
    TmpWord[0] = 0;

    for( int j=0; j < Len; j++ )
    {
        if( Direction > 0 )
            i = j;
        else
            i = Len - j - 1;

        CurChar = Line[i];
        if( IsInsideWord )
        {
            if( MyIsAlpha( CurChar ) == 0 )
            {
                IsInsideWord = FALSE;
                if( TmpWord[0] )
                {
                    if( Direction <= 0 )
                        strrev( TmpWord );
                    LineWordsCnt += AddWord( TmpWord );
                    TmpWord[0] = 0;
                };
            }
            else
            {
                CurCharStr[0] = CurChar;
                strcat( TmpWord, CurCharStr );
            };
        }
        else
            if( MyIsAlpha( CurChar ) )
            {
                IsInsideWord = TRUE;
                CurCharStr[0] = CurChar;
                strcpy( TmpWord, CurCharStr );
            }
    }
    if( TmpWord[0] )
    {
        if( Direction <= 0 )
            strrev( TmpWord );

        LineWordsCnt += AddWord( TmpWord );
    };

    return LineWordsCnt;
}

//       
//      
BOOL TCompletion::AddWord(char *NewWord)
{
    BOOL WasAdded = FALSE;

    //  NewWord==NULL  NewWord 
    if( NewWord && NewWord[0] && WordCount < WordsToFindCnt
        && (int)strlen( NewWord ) > WordLen && (int)strlen( NewWord ) >= MinWordLen )
    {
        if( AsteriskSymbol && strchr( (char *)Word, (int)AsteriskSymbol ) )
        {
            char TmpWord[MAX_WORD_LEN];
            char TmpNewWord[MAX_WORD_LEN];

            strcpy( TmpWord, (char *)Word );
            strcpy( TmpNewWord, NewWord );

            if( !CaseSensitive )
            {
                FSF.LUpperBuf( TmpWord, WordLen );
                FSF.LUpperBuf( TmpNewWord, strlen( TmpNewWord ) );
            };

            char *p = TmpWord;
            while( (p = strchr( p, AsteriskSymbol )) )
            {
                *p = '*';
                p++;
            };

            if( MatchPattern( (unsigned char *)TmpNewWord, (unsigned char *)TmpWord ) )
                WasAdded = InsertWordIntoList( NewWord );
        }
        else
        {
            if( CaseSensitive )
            {
                if( strncmp( NewWord, (char *)Word, WordLen ) == 0 )
                    WasAdded = InsertWordIntoList( NewWord );

            }
            else
            {
                if( FSF.LStrnicmp( NewWord, (char *)Word, WordLen ) == 0 )
                    WasAdded = InsertWordIntoList( NewWord );
            };
        };
    };

    return WasAdded;
};

//    
BOOL TCompletion::InsertWordIntoList(char *NewWord)
{
    BOOL WasAdded = FALSE;

    int i=0;

    for( ; i < WordCount; i++ )
        if( strcmp(NewWord, WordList[i].Text) == 0 )
        {
            //   
            (*((int *)(&WordList[i].Text[MAX_WORD_LEN])))++;
            break;
        };

    if( i == WordCount )
    {
        strcpy(WordList[WordCount].Text, NewWord);

        WordList[WordCount].Selected=FALSE;
        WordList[WordCount].Checked=FALSE;
        WordList[WordCount].Separator=FALSE;

        MaxWordLen = MAX( (unsigned int)MaxWordLen, strlen( NewWord ) );
        WordCount++;

        //   = 1 ( )
        (*((int *)(&WordList[i].Text[MAX_WORD_LEN]))) = 1;

        WasAdded = TRUE;
    };

    return WasAdded;
};

void TCompletion::GetDialogItems(void)
{
    WorkInsideWord      = DialogItems[IWorkInsideWord].Selected;
    CaseSensitive       = DialogItems[ICaseSensitive].Selected;
    BrowseDownward      = DialogItems[IBrowseDownward].Selected;
    ConsiderDigitAsChar = DialogItems[IConsiderDigitAsChar].Selected;
    PartialCompletion   = DialogItems[IPartialCompletion].Selected;
    AddTrailingSpace    = DialogItems[IAddTrailingSpace].Selected;

    BrowseLineCnt   = FSF.atoi( DialogItems[IBrowseLineCnt].Data );
    WordsToFindCnt  = FSF.atoi( DialogItems[IWordsToFindCnt].Data );
    MinWordLen      = FSF.atoi( DialogItems[IMinWordLen].Data );
    FSF.sprintf( AdditionalLetters, "%s", DialogItems[IAdditionalLetters].Data );
};

void TCompletion::SetDialogItems(void)
{
    int Msgs[]={
                  MOk, MOk, MCancel,
                  MWorkInsideWord,
                  MCaseSensitive,

                  MConsiderDigitAsChar,
                  MAdditionalLetters, MAdditionalLetters,

                  MBrowseDownward,
                  MBrowseLineCnt, MBrowseLineCnt,
                  MWordsToFindCnt, MWordsToFindCnt,
                  MMinWordLen, MMinWordLen,

                  MPartialCompletion,
                  MAddTrailingSpace,
               };
    int DialogElements[][4]={
                    {DI_DOUBLEBOX, 1,  1,  0  },
                    {DI_BUTTON,    0,  0,  0  },
                    {DI_BUTTON,    0,  0,  0  },
                    {DI_CHECKBOX,  3,  3,  0  }, // WorkInsideWord
                    {DI_CHECKBOX,  3,  4,  0  }, // CaseSensitive

                    {DI_CHECKBOX,  3,  5,  0  }, // ConsiderDigitAsChar
                    {DI_TEXT,      3,  6,  0  }, //
                    {DI_FIXEDIT,   3,  7, DialogWidth-4 },  // AdditionalLetters

                    {DI_CHECKBOX,  3,  8,  0  }, // BrowseDownward
                    {DI_TEXT,      8,  9,  0  }, //
                    {DI_FIXEDIT,   3,  9,  6  }, // BrowseLineCnt
                    {DI_TEXT,      6, 10,  0  }, //
                    {DI_FIXEDIT,   3, 10,  4  }, // WordsToFindCnt
                    {DI_TEXT,      6, 11,  0  }, //
                    {DI_FIXEDIT,   3, 11,  4  }, // MinWordLen

                    {DI_CHECKBOX,  3, 12,  0  }, // PartialCompletion
                    {DI_CHECKBOX,  3, 13,  0  }, // AddTrailingSpace

                   };

    for( unsigned int i=0; i < (sizeof(Msgs)/sizeof(Msgs[0])); i++ )
    {
        DialogItems[i].Type=DialogElements[i][0];
        DialogItems[i].X1=DialogElements[i][1];
        DialogItems[i].Y1=DialogElements[i][2];
        DialogItems[i].X2=DialogElements[i][3];
        DialogItems[i].Y2=DialogElements[i][4];
        DialogItems[i].Focus=0;
        DialogItems[i].Selected=0;
        DialogItems[i].Flags=0;
        DialogItems[i].DefaultButton=0;
        strcpy( DialogItems[i].Data, GetMsg( Msgs[i] ) ); //   - 
    };

    DialogItems[ICfg].X2=DialogWidth-2;
    DialogItems[ICfg].Y2=DialogHeight-2;

    // 
    DialogItems[IWorkInsideWord].Focus=1;
    // 
    DialogItems[IOk].Flags=DIF_CENTERGROUP;
    DialogItems[IOk].DefaultButton=TRUE;
    DialogItems[ICancel].Flags=DIF_CENTERGROUP;

    DialogItems[IOk].Y1=DialogHeight-3;

    DialogItems[ICancel].Y1=DialogHeight-3;

    DialogItems[IWorkInsideWord].Selected      = WorkInsideWord;
    DialogItems[ICaseSensitive].Selected       = CaseSensitive;
    DialogItems[IBrowseDownward].Selected      = BrowseDownward;
    DialogItems[IConsiderDigitAsChar].Selected = ConsiderDigitAsChar;
    DialogItems[IPartialCompletion].Selected   = PartialCompletion;
    DialogItems[IAddTrailingSpace].Selected    = AddTrailingSpace;

    //     
    FSF.itoa( BrowseLineCnt, DialogItems[IBrowseLineCnt].Data, 10 );
    FSF.itoa( WordsToFindCnt, DialogItems[IWordsToFindCnt].Data, 10 );
    FSF.itoa( MinWordLen, DialogItems[IMinWordLen].Data, 10 );
    FSF.sprintf( DialogItems[IAdditionalLetters].Data, "%s", AdditionalLetters );
};

void TCompletion::ShowDialog(void)
{
    DialogItems = (FarDialogItem *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, DialogItemsCnt*sizeof(FarDialogItem));
//    DialogItems = new FarDialogItem[DialogItemsCnt];

    SetDialogItems( );

    int ExitCode = 1;

    for( ; ExitCode; )
    {
        ExitCode = Info.Dialog( Info.ModuleNumber, -1, -1, DialogWidth, DialogHeight, ConfigHelpTopic, DialogItems, DialogItemsCnt );

        switch( ExitCode )
        {
            case IOk:  //   "OK"
                GetDialogItems( );
                SetOptions( );
            case ICancel:
            case -1:
                ExitCode = 0;
                break;
            default:
                ExitCode = ProcessDlgKey( ExitCode );
        };
    };

//    delete DialogItems;
    HeapFree(GetProcessHeap(), 0, DialogItems);
};

int TCompletion::ProcessDlgKey( int ItemNumber )
{
    return 0;
};
