#include "MCmpl.hpp"
#include "ECmplLng.h"

TMenuCompletion::TMenuCompletion(void)
{
    WorkInsideWord = TRUE;
    CaseSensitive = FALSE;
    BrowseDownward = TRUE;
    ConsiderDigitAsChar = TRUE;
    AddTrailingSpace = FALSE;

    BrowseLineCnt = 1000;
    WordsToFindCnt = 20;
    AdditionalLetters[0] = '_';
    AdditionalLetters[1] = 0;

    WordCount = 0;
    WordList = NULL;

    SingleVariantInMenu = FALSE;
    NotFoundSound = TRUE;
    SortListCount = 10;

    AsteriskSymbol = '*';

    RegKey[0] = 0;
    strcpy( ConfigHelpTopic, "Config" );
    strcpy( ShortCuts, "0123456789abcdefghijklmnopqrstuvwxyz" );

    DialogItemsCnt = MCMPL_DIALOG_ITEMS;

    DialogWidth = 42;
    DialogHeight = 23;
};

TMenuCompletion::~TMenuCompletion()
{
    ;
};

void TMenuCompletion::GetOptions(void)
{
    TCompletion::GetOptions( );

    SingleVariantInMenu = GetRegKey( HKEY_CURRENT_USER, RegKey, "SingleVariantInMenu", SingleVariantInMenu );
    NotFoundSound       = GetRegKey( HKEY_CURRENT_USER, RegKey, "NotFoundSound", NotFoundSound );
    AsteriskSymbol      = GetRegKey( HKEY_CURRENT_USER, RegKey, "AsteriskSymbol", AsteriskSymbol );
    int DataSize = 80;
    GetRegKey( HKEY_CURRENT_USER, RegKey, "ShortCuts", ShortCuts, ShortCuts, DataSize );

    Validate( );

    GetLocales();
};

void TMenuCompletion::SetOptions(void)
{
    TCompletion::SetOptions( );

    Validate( );

    SetRegKey( HKEY_CURRENT_USER, RegKey, "SingleVariantInMenu", SingleVariantInMenu );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "NotFoundSound", NotFoundSound );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "AsteriskSymbol", AsteriskSymbol );
    SetRegKey( HKEY_CURRENT_USER, RegKey, "ShortCuts", ShortCuts );
};

void TMenuCompletion::Validate(void)
{
    TCompletion::Validate( );

    SingleVariantInMenu = SingleVariantInMenu ? TRUE : FALSE;
    NotFoundSound       = NotFoundSound ? TRUE : FALSE;
}

int TMenuCompletion::ShowMenu(void)
{
    char BottomMsg[MAX_WORD_LEN];
    struct EditorInfo ei;

    EditorControl( ECTL_GETINFO, &ei );

    FSF.sprintf( BottomMsg, GetMsg( MHave ), WordCount );

    int MenuWidth = MAX( (unsigned int)(MaxWordLen), strlen( GetMsg( MChooseWord ) )+1 );
    MenuWidth = MAX( (unsigned int)(MenuWidth), strlen( BottomMsg )+1 );

    int CoorX=ei.CurPos-ei.LeftPos;
    int CoorY=ei.CurLine-ei.TopScreenLine;
    int WindowSizeX=ei.WindowSizeX;
    int WindowSizeY=ei.WindowSizeY;

    int MenuX = MAX( 0, CoorX+1-((signed)(WordLen))-MenuWidth-ADD_MENU_X);
    MenuX = (WindowSizeX-CoorX) > (CoorX+2-((signed)(WordLen))) ? CoorX+3 : MenuX;

    int MenuY = (WindowSizeY-CoorY-1) > CoorY+1 ?
                    ((WindowSizeY-CoorY-1) > (WordCount+ADD_MENU_Y) ?
                        CoorY+3 : (WindowSizeY+3-WordCount-ADD_MENU_Y))
                    : (CoorY+2-WordCount-ADD_MENU_Y);
    MenuY = MAX(MenuY, 2);

    MenuWidth = Info.Menu( Info.ModuleNumber, MenuX, MenuY, 0,
              FMENU_WRAPMODE, GetMsg(MChooseWord),
              BottomMsg, // bottom
              "List", // helptopic
              BreakKeys[LocaleNumber], &MenuX, // breakkeys, breakcode
              WordList, WordCount );
    if( MenuWidth >= 0 && MenuX >= 0 )
    {
        MenuY = strlen( WordList[MenuWidth].Text );
        WordList[MenuWidth].Text[MenuY] = BreakSymbols[LocaleNumber][MenuX];
        WordList[MenuWidth].Text[MenuY+1] = 0;
    };

    return MenuWidth;
};

BOOL TMenuCompletion::CompleteWord(void)
{
    struct EditorInfo ei;
    FarMenuItem TmpWordList[MAX_WORD_CNT];

    BOOL WasCompleted = FALSE;

    EditorControl(ECTL_GETINFO,&ei);

    //     .
    if( GetPreWord() && WordsToFindCnt )
    {
        WordList = TmpWordList;

        // If something was found
        if( DoSearch() )
        {
            int ItemNumber = 0;

            DoPartialCompletion( );

//            if( !PartialCompletion || DoPartialCompletion( ) )
//            {
                if( SingleVariantInMenu || WordCount > 1 )
                {
                    SortList();

                    LocaleNumber = GetStringLocale( (char *)Word );

                    AddShortCuts();

                    ItemNumber = ShowMenu();

                    DeleteShortCut( ItemNumber );
                };

                if( ItemNumber != -1 )
                {
                    strcpy( (char *)Word, WordList[ItemNumber].Text );
                    WordLen = strlen( (char *)Word );

                    PutWord( );
                    SetCurPos( WordPos + WordLen );

                    WasCompleted = TRUE;
                };
//            };
        }
        else
            if( NotFoundSound )
                MessageBeep( MB_ICONASTERISK );
    };

    MaxWordLen = 0;
    WordCount = 0;
    WordPos = 0;
    Word[0] = 0;

    return WasCompleted;
};

void TMenuCompletion::GetDialogItems(void)
{
    TCompletion::GetDialogItems( );

    SingleVariantInMenu = DialogItems[ISingleVariantInMenu].Selected;
    NotFoundSound       = DialogItems[INotFoundSound].Selected;

    AsteriskSymbol      = DialogItems[IAsteriskSymbol].Data[0];
    SortListCount       = FSF.atoi( DialogItems[ISortListCount].Data );
};

void TMenuCompletion::SetDialogItems(void)
{
    TCompletion::SetDialogItems( );

    int Msgs[]={
                  MSingleVariantInMenu,
                  MNotFoundSound,

                  MSortListCount, MSortListCount,
                  MAsteriskSymbol, MAsteriskSymbol,
               };
    int DialogElements[][4]={
                    {DI_CHECKBOX,  3, 15,  0  }, // SingleVariantInMenu
                    {DI_CHECKBOX,  3, 16,  0  }, // NotFoundSound

                    {DI_TEXT,      6, 17,  0  }, //
                    {DI_FIXEDIT,   3, 17,  4  }, // SortListCount

                    {DI_TEXT,      6, 18,  0  }, //
                    {DI_FIXEDIT,   3, 18,  3  }, // AsteriskSymbol
                   };

    for( unsigned int i=0; i < (sizeof(Msgs)/sizeof(Msgs[0])); i++ )
    {
        DialogItems[i+CMPL_DIALOG_ITEMS].Type=DialogElements[i][0];
        DialogItems[i+CMPL_DIALOG_ITEMS].X1=DialogElements[i][1];
        DialogItems[i+CMPL_DIALOG_ITEMS].Y1=DialogElements[i][2];
        DialogItems[i+CMPL_DIALOG_ITEMS].X2=DialogElements[i][3];
        DialogItems[i+CMPL_DIALOG_ITEMS].Y2=DialogElements[i][4];
        DialogItems[i+CMPL_DIALOG_ITEMS].Focus=0;
        DialogItems[i+CMPL_DIALOG_ITEMS].Selected=0;
        DialogItems[i+CMPL_DIALOG_ITEMS].Flags=0;
        DialogItems[i+CMPL_DIALOG_ITEMS].DefaultButton=0;
        strcpy( DialogItems[i+CMPL_DIALOG_ITEMS].Data, GetMsg( Msgs[i] ) ); //   - 
    };

    DialogItems[ISingleVariantInMenu].Selected = SingleVariantInMenu;
    DialogItems[INotFoundSound].Selected = NotFoundSound;

    //     
    FSF.itoa( SortListCount, DialogItems[ISortListCount].Data, 10 );
    DialogItems[IAsteriskSymbol].Data[0] = AsteriskSymbol;
    DialogItems[IAsteriskSymbol].Data[1] = 0;

    strcpy( DialogItems[ICfg].Data, GetMsg( MMenuCfg ) ); // 
};

int TMenuCompletion::AddShortCuts(void)
{
    int ShortCutCnt = MIN( WordCount, (int)strlen( ShortCuts ) );
    int i = 0;

    for( ; i < ShortCutCnt; i++ )
    {
        int ZeroPos = MIN( strlen( WordList[i].Text ), MAX_WORD_LEN-4 );
        memmove( &WordList[i].Text[3], WordList[i].Text, ZeroPos );
        WordList[i].Text[ZeroPos+3] = 0;
        WordList[i].Text[0] = '&';
        WordList[i].Text[1] = ShortCuts[i];
        WordList[i].Text[2] = ' ';
    };

    for( ; i < WordCount; i++ )
    {
        int ZeroPos = MIN( strlen( WordList[i].Text ), MAX_WORD_LEN-3 );
        memmove( &WordList[i].Text[2], WordList[i].Text, ZeroPos );
        WordList[i].Text[ZeroPos+2] = 0;
        WordList[i].Text[0] = ' ';
        WordList[i].Text[1] = ' ';
    };

    return ShortCutCnt;
};

BOOL TMenuCompletion::DeleteShortCut( int ItemNumber )
{
    int WasDeleted = FALSE;

    if( ItemNumber >= 0 && ItemNumber < WordCount )
    {
        int ShortCutCnt = MIN( WordCount, (int)strlen( ShortCuts ) );
        int Offset = ItemNumber < ShortCutCnt ? 3 : 2;

        memmove( WordList[ItemNumber].Text, &WordList[ItemNumber].Text[Offset],
                 strlen( WordList[ItemNumber].Text ) - Offset + 1 );
        WasDeleted = TRUE;
    };

    return WasDeleted;
};


int TMenuCompletion::ProcessDlgKey( int ItemNumber )
{
    return 0;
};
